/*
 * signal.c
 *
 * Copyright 2004, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 */


#include<types.h>
#include<lib.h>
#include<proc.h>
#include<interrupt.h>
#include<mm.h>
#include<lock.h>
#include<errno.h>
#include<term.h>
#include<signal.h>
#include<test.h>


/****************************************************************************************
 *
 * < signal default actions >
 *
 *****************************************************************************************/

/*
 * PRIVATE
 * Default aactinos.
 */
/* terminate process. */
static void terminate(int signum)
{
	exit(0,signum);
}

/* abnormal terminate process. */
static void abnormalTerminate(int signum)
{
	exit(0,signum);
}

/* stop process. */
static void stopProcess(int signum)
{
	PROC *proc=get_current_task();


	proc->signum=signum;
/**************************************************************/
	/* ƥץwaitƤ鵯 */
	if((isForegrouond(proc))&&(proc->parent->state&TASK_CHILD_WAIT))
			add_to_schedule(proc->parent,TASK_SIGNAL_WAIT);
/***************************************************************/
	del_from_schedule(TASK_SIGNAL_WAIT);
	wait_task();
}


/*
 * PUBLIC
 * signal actions.
 */
static SIG_ACTION defaultAction[]={
	NULL,
	abnormalTerminate,	/* SIGABRT		A */
	terminate,			/* SIGALRM		T */
	abnormalTerminate,	/* SIGBUS		A */
	SIG_IGN,			/* SIGCHLD		I */
	SIG_DFL,			/* SIGCONT		C */
	abnormalTerminate,	/* SIGFPE		A */
	terminate,			/* SIGHUP		T */
	abnormalTerminate,	/* SIGILL		A */
	terminate,			/* SIGINT		T */
	terminate,			/* SIGKILL		T */
	terminate,			/* SIGPIPE		T */
	terminate,			/* SIGPOLL		T */
	terminate,			/* SIGPROF		T */
	abnormalTerminate,	/* SIGQUIT		A */
	abnormalTerminate,	/* SIGSEGV		A */
	stopProcess,		/* SIGSTOP		S */
	abnormalTerminate,	/* SIGSYS		A */
	terminate,			/* SIGTERM		T */
	abnormalTerminate,	/* SIGTRAP		A */
	stopProcess,		/* SIGTSTP		S */
	stopProcess,		/* SIGTTIN		S */
	stopProcess,		/* SIGTTOU		S */
	SIG_IGN,			/* SIGURG		I */
	terminate,			/* SIGUSR1		T */
	terminate,			/* SIGUSR2		T */
	terminate,			/* SIGVTALRM	T */
	abnormalTerminate,	/* SIGXCPU		A */
	abnormalTerminate,	/* SIGXFSZ		A */
};


/****************************************************************************************
 *
 * < ʥ >
 *
 *****************************************************************************************/


/*
 * GLOBALk 
 * ʥ¹ԡ
 * parameters : user esp address
 */
void doSignal(uint *user_esp)
{
	SIGNAL *sig_st;
	PROC *proc=get_current_task();
	int i;


	if(proc->nest_count>0)return;

	sig_st=proc->signal_struct;

	/* ʥϥɥμ¹ԡ */
	while(sig_st->signal!=0)
	{
		for(i=0;(sig_st->signal&(1<<i))==0;++i);
		sig_st->signal^=(1<<i);		/* ʥ륯ꥢ*/

		if((uint)sig_st->action[i]>=USER_BEG)
		{
			uint before_mask;

			/* ʥޥ촹 */
			before_mask=sig_st->mask;
			sig_st->mask|=sig_st->actMask[i]|(1<<i);

			/* 桼ϥɥ¹ԡ */
			doUserHandler(i,sig_st->action[i],*user_esp,&sig_st->beforeEsp,&sig_st->beforeStackEntry);
			sig_st->beforeEsp=0;

			/* ʥޥ᤹ */
			sig_st->mask=before_mask;
		}
		else if(sig_st->action[i]!=SIG_DFL)
		{
			/* ǥեȥϥɥ¹ԡ */
			sig_st->action[i](i);
		}
	}
}


/*
 * GLOBAL
 * ʥ
 * ǥեȥΡcontinue processפϡʥץ
 * ư롣
 * parameters : process,ʥ
 * return : 0 or error number
 */
int sendSignal(PROC *proc,int signal)
{
	int sig;
	SIGNAL *sig_st;

	
	sig_st=proc->signal_struct;

	/* ʥ뤬̵ˤƤʤ */
	if(sig_st->action[signal]==SIG_IGN)return -EPERM;

	sig_st->reserve|=(1<<signal)&sig_st->mask;		/* αΥʥ롣 */
	sig=((1<<signal)|sig_st->mask)^sig_st->mask;	/* αʤʥ롣 */

	if(sig!=0)
	{
		enter_spinlock(&sig_st->lock);
		{
			sig_st->signal|=sig;
			if(proc->state==TASK_SIGNAL_WAIT)
			{
				add_to_schedule(proc,TASK_SIGNAL_WAIT);
				sig_st->signal&=~(1<<SIGCONT);					/* SIGCONTϥꥢ */
			}
		}
		exit_spinlock(&sig_st->lock);
	}

	return 0;
}


/*
 * GLOBAL
 * ʥ
 * parameters : process,ʥ
 * return : 0 or error=-1
 */
int forceSendSignal(PROC *proc,int signal)
{
	SIGNAL *sig = proc->signal_struct;


	enter_spinlock(&sig->lock);
	{
		sig->signal |= 1 << signal;
		if(proc->state == TASK_SIGNAL_WAIT)
		{
			add_to_schedule(proc,TASK_SIGNAL_WAIT);
			sig->signal &= ~(1 << SIGCONT);			/* SIGCONTϥꥢ */
		}
	}
	exit_spinlock(&sig->lock);

	return 0;
}


/*
 * GLOBAL
 * ʥԤ
 */
void sigWait()
{
	SIGNAL *sig=get_current_task()->signal_struct;


	enter_spinlock(&sig->lock);
	{
		if(sig->signal!=0)
		{
			exit_spinlock(&sig->lock);
			
			sig->signal&=~(1<<SIGCONT);					/* SIGCONTϥꥢ */
			return ;
		}

		cli();
		del_from_schedule(TASK_SIGNAL_WAIT);
	}
	exit_spinlock(&sig->lock);
	sti();

	wait_task();
}


/*
 * GLOBAL
 * fork()Υʥ뵡ꡣ
 * parameters : ƥץΥʥ빽¤Ρ
 * return : 0 or error number
 */
SIGNAL *setSignal(SIGNAL *parent_sig)
{
	SIGNAL *sig_st;


	if((sig_st=kmalloc(sizeof(SIGNAL)))==NULL)return NULL;

	memcpy(sig_st,parent_sig,sizeof(SIGNAL));
	sig_st->signal=0;
	sig_st->reserve=0;

	return sig_st;
}


/*
 * GLOBAL
 * exit()Υʥ뵡γ
 * parameters : ʥ빽¤
 */
void releaseSignal(SIGNAL *sig_st)
{
	kfree(sig_st);
}


/*
 * GLOBAL
 * ʥ뵡Υꥻåȡ
 * parameters : ʥ빽¤
 */
void resetSignal(SIGNAL *sig_st)
{
	/* ǥեȥ򥳥ԡ */
	memcpy(sig_st->action,defaultAction,sizeof(SIG_ACTION)*SIG_ACTIONS);

	memset(sig_st->actMask,0,sizeof(uint)*SIG_ACTIONS);
	sig_st->signal=0;
	sig_st->reserve=0;
	sig_st->beforeEsp=0;
}


/*
 * GLOBAL
 * espͤ롣
 */
uint getBeforeEsp()
{
	return ((SIGNAL*)get_current_task()->signal_struct)->beforeEsp;
}


/*
 * GLOBAL
 * Υͥ륹åʪɥ쥹롣
 */
uint getBeforeStackEntry()
{
	return ((SIGNAL*)get_current_task()->signal_struct)->beforeStackEntry;
}


/*
 * GLOBAL
 * SIGINTȯ
 * return : ȯ=0ʳ
 */
int isSigint()
{
	return ((SIGNAL*)get_current_task()->signal_struct)->signal&1<<SIGINT;
}


/*
 * GLOBAL
 * ʥ뵡ν
 * return : SIGNAL struct or NULL
 */
SIGNAL *initSignal()
{
	SIGNAL *sig_st;


	if((sig_st=kmalloc(sizeof(SIGNAL)))==NULL)return NULL;

	/*  */
	resetSignal(sig_st);	
	sig_st->mask=0;
	sig_st->lock=0;

	return sig_st;
}


/****************************************************************************************
 *
 * < ƥॳ >
 *
 *****************************************************************************************/

int sys_sigaction(int signum,const struct sigaction *act,struct sigaction *oldact)
{
	SIGNAL *signal;
	struct sigaction tmp_act;


	if(signum==0)return 0;
	if((signum>=SIG_ACTIONS)||(signum<0))return -EINVAL;

	signal=(SIGNAL*)get_current_task()->signal_struct;
	tmp_act=*act;

	if(oldact!=NULL)
	{
		if(signal->action[signum]==SIG_IGN)oldact->sa_handler=SIG_IGN;
		else if(signal->action[signum]==defaultAction[signum])oldact->sa_handler=SIG_DFL;
		else oldact->sa_handler=signal->action[signum];

		oldact->sa_mask=signal->actMask[signum];
	}
	if(act!=NULL)
	{
		if((signum==SIGKILL)||(signum==SIGSTOP))return -EINVAL;		/* SIGKILLSIGSTOPѹԲġ */


		if(tmp_act.sa_handler==SIG_DFL)signal->action[signum]=defaultAction[signum];
		else if(tmp_act.sa_handler==SIG_IGN)signal->action[signum]=SIG_IGN;
		else if((uint)tmp_act.sa_handler>=USER_BEG)signal->action[signum]=tmp_act.sa_handler;
		else return -EINVAL;		/* ϥɥ饢ɥ쥹 */

		signal->actMask[signum]=tmp_act.sa_mask;
	}

	return 0;
}


int sys_sigprocmask(int how,uint *set,uint *oldset)
{
	uint tmp_mask=*set;
	SIGNAL *sig_st;


	sig_st=get_current_task()->signal_struct;

	if(oldset!=NULL)*oldset=sig_st->mask;

	switch(how)
	{
		case SIG_BLOCK:
			if(tmp_mask&((1<<SIGKILL)|(1<<SIGSTOP)))return -EINVAL;	/* SIGKILLSIGSTOPϥ֥åԲġ */
			sig_st->mask|=tmp_mask;
			break;
		case SIG_UNBLOCK:
			sig_st->mask&=~tmp_mask;
			break;
		case SIG_SETMASK:
			if(tmp_mask&((1<<SIGKILL)|(1<<SIGSTOP)))return -EINVAL;	/* SIGKILLSIGSTOPϥ֥åԲġ */
			sig_st->mask=tmp_mask;
			break;
	}

	return 0;
}


int sys_kill(pid_t pid,int sig)
{
	pid_t pgid;
	int rest;
	PROC *proc;


	if((uint)sig>=SIG_ACTIONS)return -EPERM;

	if(pid>0)
	{
		if((proc=searchProcList(pid))==NULL)return -ESRCH;
		rest=sendSignal(proc,sig);
	}
	else if(pid==0)		/* callץƱ롼ס */
	{
		rest=-ESRCH;
		proc=get_current_task();
		pgid=proc->pgid;
		while((proc=getNextProc(proc,pgid))!=NULL)
			if((rest=sendSignal(proc,sig))<0)break;
	}
	else if(pid==-1)	/* ץ */
	{
		rest=-ESRCH;
		for(proc=NULL;(proc=getAllProc(proc))!=NULL;)
			if((rest=sendSignal(proc,sig))<0)break;
	}
	else				/* pid<-1 -pidΥ롼ס */
	{
		rest=-ESRCH;
		pgid=-pid;
		for(proc=NULL;(proc=getNextProc(proc,pgid))!=NULL;)
			if((rest=sendSignal(proc,sig))<0)break;
	}

	return rest;
}


int sys_sigpending(sigset_t *set,int flag)
{
	SIGNAL *sig_st;


	sig_st=get_current_task()->signal_struct;

	switch(flag)
	{
		case SIG_PENDING_SOTORE:
			if(checkMem(set)==-1)return -EINVAL;
			*set=sig_st->reserve;
			break;
		case SIG_PENDING_SET:
			sig_st->reserve=*set;
			break;
	}

	return 0;
}


int sys_sigsuspend(const sigset_t *mask)
{
	uint before_mask;
	SIGNAL *sig_st;


	/* SIGKILLSIGSTOPϥ֥åԲġ */
	if(*mask&((1<<SIGKILL)|(1<<SIGSTOP)))return -EINVAL;

	sig_st=get_current_task()->signal_struct;
	before_mask=sig_st->mask;
	sig_st->mask=*mask;

	/* suspend. */
	del_from_schedule(TASK_SIGNAL_WAIT);
	wait_task();

	sig_st->mask=before_mask;

	return -EINTR;
}
